@@ -1,7 +1,21 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
-# Copyright (C) 2006-2007 Lukáš Lalinský
+#
+# Copyright (C) 2006-2009, 2011-2012 Lukáš Lalinský
+# Copyright (C) 2008-2011, 2014, 2018-2020 Philipp Wolfer
+# Copyright (C) 2009 Carlin Mangar
+# Copyright (C) 2011-2012 Johannes Weißl
+# Copyright (C) 2011-2014 Michael Wiencek
+# Copyright (C) 2011-2014 Wieland Hoffmann
+# Copyright (C) 2013 Calvin Walton
+# Copyright (C) 2013-2014, 2017-2019 Laurent Monin
+# Copyright (C) 2013-2015, 2017 Sophist-UK
+# Copyright (C) 2015 Frederik “Freso” S. Olesen
+# Copyright (C) 2016 Christoph Reiter
+# Copyright (C) 2016-2018 Sambhav Kothari
+# Copyright (C) 2017 tungol
+# Copyright (C) 2019 Zenara Daley
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -17,42 +31,60 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
+from collections import defaultdict
+import re
+from urllib.parse import urlparse
+
+from mutagen import id3
+import mutagen.aiff
 import mutagen.apev2
 import mutagen.dsf
 import mutagen.mp3
 import mutagen.trueaudio
-import mutagen.aiff
 
-import re
-from collections import defaultdict
-from mutagen import id3
-from picard import config, log
-from picard.coverart.image import TagCoverArtImage, CoverArtImageError
-from picard.metadata import Metadata
+from picard import (
+    config,
+    log,
+)
+from picard.coverart.image import (
+    CoverArtImageError,
+    TagCoverArtImage,
+)
 from picard.file import File
-from picard.formats.mutagenext import compatid3
-from picard.util import encode_filename, sanitize_date
-from urllib.parse import urlparse
+from picard.formats.mutagenext import (
+    compatid3,
+    delall_ci,
+)
+from picard.metadata import Metadata
+from picard.util import (
+    encode_filename,
+    sanitize_date,
+)
+from picard.util.tags import parse_comment_tag
 
 
+id3.GRP1 = compatid3.GRP1
 id3.TCMP = compatid3.TCMP
 id3.TSO2 = compatid3.TSO2
 id3.TSOC = compatid3.TSOC
 
-__ID3_IMAGE_TYPE_MAP = {
-    "other": 0,
-    "obi": 0,
-    "tray": 0,
-    "spine": 0,
-    "sticker": 0,
-    "front": 3,
-    "back": 4,
-    "booklet": 5,
-    "medium": 6,
-    "track": 6,
-}
+__IMAGE_TYPES = [
+    ("obi", 0),
+    ("tray", 0),
+    ("spine", 0),
+    ("sticker", 0),
+    ("other", 0),
+    ("front", 3),
+    ("back", 4),
+    ("booklet", 5),
+    ("track", 6),
+    ("medium", 6),
+]
 
-__ID3_REVERSE_IMAGE_TYPE_MAP = dict([(v, k) for k, v in __ID3_IMAGE_TYPE_MAP.items()])
+__ID3_IMAGE_TYPE_MAP = dict(__IMAGE_TYPES)
+
+__ID3_REVERSE_IMAGE_TYPE_MAP = dict([(v, k) for k, v in __IMAGE_TYPES])
 
 
 def id3text(text, encoding):
@@ -74,7 +106,15 @@ def image_type_as_id3_num(texttype):
 
 
 def types_from_id3(id3type):
-    return [string_(image_type_from_id3_num(id3type))]
+    return [image_type_from_id3_num(id3type)]
+
+
+def _remove_people_with_role(tags, frames, role):
+    for frame in tags.values():
+        if frame.FrameID in frames:
+            for people in list(frame.people):
+                if people[0] == role:
+                    frame.people.remove(people)
 
 
 class ID3File(File):
@@ -91,7 +131,7 @@ class ID3File(File):
 
     __translate = {
         # In same sequence as defined at http://id3.org/id3v2.4.0-frames
-        'TIT1': 'grouping',
+        # 'TIT1': 'grouping', # Depends on itunes_compatible_grouping
         'TIT2': 'title',
         'TIT3': 'subtitle',
         'TALB': 'album',
@@ -123,13 +163,16 @@ class ID3File(File):
         'COMM': 'comment',
         'TOAL': 'originalalbum',
         'TOPE': 'originalartist',
+        'TOFN': 'originalfilename',
 
         # The following are informal iTunes extensions to id3v2:
         'TCMP': 'compilation',
         'TSOC': 'composersort',
         'TSO2': 'albumartistsort',
+        'MVNM': 'movement'
     }
     __rtranslate = dict([(v, k) for k, v in __translate.items()])
+    __translate['GRP1'] = 'grouping'  # Always read, but writing depends on itunes_compatible_grouping
 
     __translate_freetext = {
         'MusicBrainz Artist Id': 'musicbrainz_artistid',
@@ -154,13 +197,33 @@ class ID3File(File):
         'BARCODE': 'barcode',
         'ASIN': 'asin',
         'MusicMagic Fingerprint': 'musicip_fingerprint',
-        'Artists': 'artists',
-        'Work': 'work',
+        'ARTISTS': 'artists',
+        'WORK': 'work',
         'Writer': 'writer',
+        'SHOWMOVEMENT': 'showmovement',
     }
     __rtranslate_freetext = dict([(v, k) for k, v in __translate_freetext.items()])
     __translate_freetext['writer'] = 'writer'  # For backward compatibility of case
 
+    # Freetext fields that are loaded case-insensitive
+    __rtranslate_freetext_ci = {
+        'replaygain_album_gain': 'REPLAYGAIN_ALBUM_GAIN',
+        'replaygain_album_peak': 'REPLAYGAIN_ALBUM_PEAK',
+        'replaygain_album_range': 'REPLAYGAIN_ALBUM_RANGE',
+        'replaygain_track_gain': 'REPLAYGAIN_TRACK_GAIN',
+        'replaygain_track_peak': 'REPLAYGAIN_TRACK_PEAK',
+        'replaygain_track_range': 'REPLAYGAIN_TRACK_RANGE',
+        'replaygain_reference_loudness': 'REPLAYGAIN_REFERENCE_LOUDNESS',
+    }
+    __translate_freetext_ci = dict([(b.lower(), a) for a, b in __rtranslate_freetext_ci.items()])
+
+    # Obsolete tag names which will still be loaded, but will get renamed on saving
+    __rename_freetext = {
+        'Artists': 'ARTISTS',
+        'Work': 'WORK',
+    }
+    __rrename_freetext = dict([(v, k) for k, v in __rename_freetext.items()])
+
     _tipl_roles = {
         'engineer': 'engineer',
         'arranger': 'arranger',
@@ -171,12 +234,18 @@ class ID3File(File):
     _rtipl_roles = dict([(v, k) for k, v in _tipl_roles.items()])
 
     __other_supported_tags = ("discnumber", "tracknumber",
-                              "totaldiscs", "totaltracks")
+                              "totaldiscs", "totaltracks",
+                              "movementnumber", "movementtotal")
     __tag_re_parse = {
         'TRCK': re.compile(r'^(?P<tracknumber>\d+)(?:/(?P<totaltracks>\d+))?$'),
-        'TPOS': re.compile(r'^(?P<discnumber>\d+)(?:/(?P<totaldiscs>\d+))?$')
+        'TPOS': re.compile(r'^(?P<discnumber>\d+)(?:/(?P<totaldiscs>\d+))?$'),
+        'MVIN': re.compile(r'^(?P<movementnumber>\d+)(?:/(?P<movementtotal>\d+))?$')
     }
 
+    def __init__(self, filename):
+        super().__init__(filename)
+        self.__casemap = {}
+
     def build_TXXX(self, encoding, desc, values):
         """Construct and return a TXXX frame."""
         # This is here so that plugins can customize the behavior of TXXX
@@ -189,6 +258,7 @@ class ID3File(File):
 
     def _load(self, filename):
         log.debug("Loading file %r", filename)
+        self.__casemap = {}
         file = self._get_file(encode_filename(filename))
         tags = file.tags or {}
         # upgrade custom 2.3 frames to 2.4
@@ -201,16 +271,26 @@ class ID3File(File):
             frameid = frame.FrameID
             if frameid in self.__translate:
                 name = self.__translate[frameid]
-                if frameid.startswith('T'):
+                if frameid.startswith('T') or frameid in ["GRP1", "MVNM"]:
                     for text in frame.text:
                         if text:
-                            metadata.add(name, string_(text))
+                            metadata.add(name, text)
                 elif frameid == 'COMM':
                     for text in frame.text:
                         if text:
-                            metadata.add('%s:%s' % (name, frame.desc), string_(text))
+                            if frame.lang == 'eng':
+                                name = '%s:%s' % (name, frame.desc)
+                            else:
+                                name = '%s:%s:%s' % (name, frame.lang, frame.desc)
+                            metadata.add(name, text)
                 else:
-                    metadata.add(name, string_(frame))
+                    metadata.add(name, frame)
+            elif frameid == 'TIT1':
+                itunes_compatible = config.setting['itunes_compatible_grouping']
+                name = 'work' if itunes_compatible else 'grouping'
+                for text in frame.text:
+                    if text:
+                        metadata.add(name, text)
             elif frameid == "TMCL":
                 for role, name in frame.people:
                     if role or name:
@@ -225,10 +305,17 @@ class ID3File(File):
                         metadata.add('performer:%s' % role, name)
             elif frameid == 'TXXX':
                 name = frame.desc
-                if name in self.__translate_freetext:
+                name_lower = name.lower()
+                if name in self.__rename_freetext:
+                    name = self.__rename_freetext[name]
+                if name_lower in self.__translate_freetext_ci:
+                    orig_name = name
+                    name = self.__translate_freetext_ci[name_lower]
+                    self.__casemap[name] = orig_name
+                elif name in self.__translate_freetext:
                     name = self.__translate_freetext[name]
-                elif ((name in self.__rtranslate) !=
-                        (name in self.__rtranslate_freetext)):
+                elif ((name in self.__rtranslate)
+                      != (name in self.__rtranslate_freetext)):
                     # If the desc of a TXXX frame conflicts with the name of a
                     # Picard tag, load it into ~id3:TXXX:desc rather than desc.
                     #
@@ -238,12 +325,12 @@ class ID3File(File):
                     # ways.) Currently, the only tag in both is license.
                     name = '~id3:TXXX:' + name
                 for text in frame.text:
-                    metadata.add(name, string_(text))
+                    metadata.add(name, text)
             elif frameid == 'USLT':
                 name = 'lyrics'
                 if frame.desc:
                     name += ':%s' % frame.desc
-                metadata.add(name, string_(frame.text))
+                metadata.add(name, frame.text)
             elif frameid == 'UFID' and frame.owner == 'http://musicbrainz.org':
                 metadata['musicbrainz_recordingid'] = frame.data.decode('ascii', 'ignore')
             elif frameid in self.__tag_re_parse.keys():
@@ -267,11 +354,11 @@ class ID3File(File):
                 except CoverArtImageError as e:
                     log.error('Cannot load image from %r: %s' % (filename, e))
                 else:
-                    metadata.append_image(coverartimage)
+                    metadata.images.append(coverartimage)
             elif frameid == 'POPM':
                 # Rating in ID3 ranges from 0 to 255, normalize this to the range 0 to 5
                 if frame.email == config.setting['rating_user_email']:
-                    rating = string_(int(round(frame.rating / 255.0 * (config.setting['rating_steps'] - 1))))
+                    rating = int(round(frame.rating / 255.0 * (config.setting['rating_steps'] - 1)))
                     metadata.add('~rating', rating)
 
         if 'date' in metadata:
@@ -289,7 +376,8 @@ class ID3File(File):
 
         if config.setting['clear_existing_tags']:
             tags.clear()
-        if metadata.images_to_be_saved_to_tags:
+        images_to_save = list(metadata.images.to_be_saved_to_tags())
+        if images_to_save:
             tags.delall('APIC')
 
         encoding = {'utf-8': 3, 'utf-16': 1}.get(config.setting['id3v2_encoding'], 0)
@@ -308,12 +396,19 @@ class ID3File(File):
                 text = metadata['discnumber']
             tags.add(id3.TPOS(encoding=0, text=id3text(text, 0)))
 
+        if 'movementnumber' in metadata:
+            if 'movementtotal' in metadata:
+                text = '%s/%s' % (metadata['movementnumber'], metadata['movementtotal'])
+            else:
+                text = metadata['movementnumber']
+            tags.add(id3.MVIN(encoding=0, text=id3text(text, 0)))
+
         # This is necessary because mutagens HashKey for APIC frames only
         # includes the FrameID (APIC) and description - it's basically
         # impossible to save two images, even of different types, without
         # any description.
         counters = defaultdict(lambda: 0)
-        for image in metadata.images_to_be_saved_to_tags:
+        for image in images_to_save:
             desc = desctag = image.comment
             if counters[desc] > 0:
                 if desc:
@@ -329,23 +424,28 @@ class ID3File(File):
 
         tmcl = mutagen.id3.TMCL(encoding=encoding, people=[])
         tipl = mutagen.id3.TIPL(encoding=encoding, people=[])
-
-        tags.delall('TCMP')
         for name, values in metadata.rawitems():
             values = [id3text(v, encoding) for v in values]
             name = id3text(name, encoding)
+            name_lower = name.lower()
 
-            if name.startswith('performer:'):
+            if not self.supports_tag(name):
+                continue
+            elif name.startswith('performer:'):
                 role = name.split(':', 1)[1]
                 for value in values:
-                    tmcl.people.append([role, value])
-            elif name.startswith('comment:'):
-                desc = name.split(':', 1)[1]
+                    if config.setting['write_id3v23']:
+                        # TIPL will be upgraded to IPLS
+                        tipl.people.append([role, value])
+                    else:
+                        tmcl.people.append([role, value])
+            elif name == 'comment' or name.startswith('comment:'):
+                (lang, desc) = parse_comment_tag(name)
                 if desc.lower()[:4] == 'itun':
                     tags.delall('COMM:' + desc)
                     tags.add(id3.COMM(encoding=0, desc=desc, lang='eng', text=[v + '\x00' for v in values]))
                 else:
-                    tags.add(id3.COMM(encoding=encoding, desc=desc, lang='eng', text=values))
+                    tags.add(id3.COMM(encoding=encoding, desc=desc, lang=lang, text=values))
             elif name.startswith('lyrics:') or name == 'lyrics':
                 if ':' in name:
                     desc = name.split(':', 1)[1]
@@ -370,6 +470,15 @@ class ID3File(File):
                 # Convert rating to range between 0 and 255
                 rating = int(round(float(values[0]) * 255 / (config.setting['rating_steps'] - 1)))
                 tags.add(id3.POPM(email=config.setting['rating_user_email'], rating=rating, count=count))
+            elif name == 'grouping':
+                if config.setting['itunes_compatible_grouping']:
+                    tags.add(id3.GRP1(encoding=encoding, text=values))
+                else:
+                    tags.add(id3.TIT1(encoding=encoding, text=values))
+            elif name == 'work' and config.setting['itunes_compatible_grouping']:
+                tags.add(id3.TIT1(encoding=encoding, text=values))
+                tags.delall('TXXX:Work')
+                tags.delall('TXXX:WORK')
             elif name in self.__rtranslate:
                 frameid = self.__rtranslate[name]
                 if frameid.startswith('W'):
@@ -383,7 +492,7 @@ class ID3File(File):
                     elif frameid == 'WOAR' and valid_urls:
                         for url in values:
                             tags.add(id3.WOAR(url=url))
-                elif frameid.startswith('T'):
+                elif frameid.startswith('T') or frameid == 'MVNM':
                     if config.setting['write_id3v23']:
                         if frameid == 'TMOO':
                             tags.add(self.build_TXXX(encoding, 'mood', values))
@@ -396,8 +505,18 @@ class ID3File(File):
                         tags.delall('XSOP')
                     elif frameid == 'TSO2':
                         tags.delall('TXXX:ALBUMARTISTSORT')
+            elif name_lower in self.__rtranslate_freetext_ci:
+                if name_lower in self.__casemap:
+                    description = self.__casemap[name_lower]
+                else:
+                    description = self.__rtranslate_freetext_ci[name_lower]
+                delall_ci(tags, 'TXXX:' + description)
+                tags.add(self.build_TXXX(encoding, description, values))
             elif name in self.__rtranslate_freetext:
-                tags.add(self.build_TXXX(encoding, self.__rtranslate_freetext[name], values))
+                description = self.__rtranslate_freetext[name]
+                if description in self.__rrename_freetext:
+                    tags.delall('TXXX:' + self.__rrename_freetext[description])
+                tags.add(self.build_TXXX(encoding, description, values))
             elif name.startswith('~id3:'):
                 name = name[5:]
                 if name.startswith('TXXX:'):
@@ -410,8 +529,10 @@ class ID3File(File):
             elif not name.startswith("~") and name not in self.__other_supported_tags:
                 tags.add(self.build_TXXX(encoding, name, values))
 
-        tags.add(tmcl)
-        tags.add(tipl)
+        if tmcl.people:
+            tags.add(tmcl)
+        if tipl.people:
+            tags.add(tipl)
 
         self._remove_deleted_tags(metadata, tags)
 
@@ -420,7 +541,7 @@ class ID3File(File):
         if self._IsMP3 and config.setting["remove_ape_from_mp3"]:
             try:
                 mutagen.apev2.delete(encode_filename(filename))
-            except:
+            except BaseException:
                 pass
 
     def _remove_deleted_tags(self, metadata, tags):
@@ -430,55 +551,46 @@ class ID3File(File):
             try:
                 if name.startswith('performer:'):
                     role = name.split(':', 1)[1]
-                    for key, frame in tags.items():
-                        if frame.FrameID in ('TMCL', 'TIPL', 'IPLS'):
-                            for people in frame.people:
-                                if people[0] == role:
-                                    frame.people.remove(people)
-                elif name.startswith('comment:'):
-                    desc = name.split(':', 1)[1]
-                    if desc.lower()[:4] != 'itun':
-                        for key, frame in list(tags.items()):
-                            if frame.FrameID == 'COMM' and frame.desc == desc:
-                                del tags[key]
+                    _remove_people_with_role(tags, ['TMCL', 'TIPL', 'IPLS'], role)
+                elif name.startswith('comment:') or name == 'comment':
+                    (lang, desc) = parse_comment_tag(name)
+                    for key, frame in list(tags.items()):
+                        if (frame.FrameID == 'COMM' and frame.desc == desc
+                            and frame.lang == lang):
+                            del tags[key]
                 elif name.startswith('lyrics:') or name == 'lyrics':
                     if ':' in name:
                         desc = name.split(':', 1)[1]
                     else:
                         desc = ''
                     for key, frame in list(tags.items()):
-                        if frame.FrameID == desc:
+                        if frame.FrameID == 'USLT' and frame.desc == desc:
                             del tags[key]
                 elif name in self._rtipl_roles:
                     role = self._rtipl_roles[name]
-                    for key, frame in tags.items():
-                        if frame.FrameID in ('TIPL', 'IPLS'):
-                            for people in frame.people:
-                                if people[0] == role:
-                                    frame.people.remove(people)
+                    _remove_people_with_role(tags, ['TIPL', 'IPLS'], role)
                 elif name == 'musicbrainz_recordingid':
                     for key, frame in list(tags.items()):
                         if frame.FrameID == 'UFID' and frame.owner == 'http://musicbrainz.org':
                             del tags[key]
                 elif real_name == 'POPM':
+                    user_email = config.setting['rating_user_email']
                     for key, frame in list(tags.items()):
-                        if frame.FrameID == 'POPM' and frame.email == config.setting['rating_user_email']:
+                        if frame.FrameID == 'POPM' and frame.email == user_email:
                             del tags[key]
                 elif real_name in self.__translate:
                     del tags[real_name]
+                elif name.lower() in self.__rtranslate_freetext_ci:
+                    delall_ci(tags, 'TXXX:' + self.__rtranslate_freetext_ci[name.lower()])
                 elif real_name in self.__translate_freetext:
-                    for key, frame in list(tags.items()):
-                        if frame.FrameID == 'TXXX' and frame.desc == real_name:
-                            del tags[key]
-                elif not name.startswith("~") and name not in self.__other_supported_tags:
-                    for key, frame in list(tags.items()):
-                        if frame.FrameID == 'TXXX' and frame.desc == name:
-                            del tags[key]
-                elif name.startswith("~"):
-                    name = name[1:]
-                    for key, frame in list(tags.items()):
-                        if frame.FrameID == 'TXXX' and frame.desc == name:
-                            del tags[key]
+                    tags.delall('TXXX:' + real_name)
+                    if real_name in self.__rrename_freetext:
+                        tags.delall('TXXX:' + self.__rrename_freetext[real_name])
+                elif not name.startswith("~id3:") and name not in self.__other_supported_tags:
+                    tags.delall('TXXX:' + name)
+                elif name.startswith("~id3:"):
+                    frameid = name[5:]
+                    tags.delall(frameid)
                 elif name in self.__other_supported_tags:
                     del tags[real_name]
             except KeyError:
@@ -486,7 +598,8 @@ class ID3File(File):
 
     @classmethod
     def supports_tag(cls, name):
-        return ((name and not name.startswith("~"))
+        unsupported_tags = ['r128_album_gain', 'r128_track_gain']
+        return ((name and not name.startswith("~") and name not in unsupported_tags)
                 or name in ("~rating", "~length")
                 or name.startswith("~id3"))
 
@@ -501,6 +614,8 @@ class ID3File(File):
             return 'TRCK'
         elif name == 'discnumber':
             return 'TPOS'
+        elif name == 'movementnumber':
+            return 'MVIN'
         else:
             return None
 
@@ -542,7 +657,6 @@ class ID3File(File):
             # ID3v23 can only save TDOR dates in YYYY format. Mutagen cannot
             # handle ID3v23 dates which are YYYY-MM rather than YYYY or
             # YYYY-MM-DD.
-
             if name == "originaldate":
                 values = [v[:4] for v in values]
             elif name == "date":
@@ -550,7 +664,6 @@ class ID3File(File):
 
             # If this is a multi-valued field, then it needs to be flattened,
             # unless it's TIPL or TMCL which can still be multi-valued.
-
             if (len(values) > 1 and name not in ID3File._rtipl_roles
                     and not name.startswith("performer:")):
                 values = [join_with.join(values)]
@@ -569,7 +682,7 @@ class MP3File(ID3File):
     _File = mutagen.mp3.MP3
 
     def _get_file(self, filename):
-        return mutagen.mp3.MP3(filename, ID3=compatid3.CompatID3)
+        return self._File(filename, ID3=compatid3.CompatID3)
 
     def _info(self, metadata, file):
         super()._info(metadata, file)
@@ -587,11 +700,7 @@ class TrueAudioFile(ID3File):
     _File = mutagen.trueaudio.TrueAudio
 
     def _get_file(self, filename):
-        return mutagen.trueaudio.TrueAudio(filename, ID3=compatid3.CompatID3)
-
-    def _info(self, metadata, file):
-        super()._info(metadata, file)
-        metadata['~format'] = self.NAME
+        return self._File(filename, ID3=compatid3.CompatID3)
 
 
 class DSFFile(ID3File):
@@ -602,11 +711,7 @@ class DSFFile(ID3File):
     _File = mutagen.dsf.DSF
 
     def _get_file(self, filename):
-        return mutagen.dsf.DSF(filename)
-
-    def _info(self, metadata, file):
-        super()._info(metadata, file)
-        metadata['~format'] = self.NAME
+        return self._File(filename, known_frames=compatid3.known_frames)
 
     def _get_tags(self, filename):
         file = self._get_file(filename)
@@ -638,6 +743,3 @@ class AiffFile(DSFFile):
     EXTENSIONS = [".aiff", ".aif", ".aifc"]
     NAME = "Audio Interchange File Format (AIFF)"
     _File = mutagen.aiff.AIFF
-
-    def _get_file(self, filename):
-        return mutagen.aiff.AIFF(filename)
