@@ -1,7 +1,26 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
-# Copyright (C) 2006 Lukáš Lalinský
+#
+# Copyright (C) 2006-2008, 2011-2012 Lukáš Lalinský
+# Copyright (C) 2007 Robert Kaye
+# Copyright (C) 2008 Gary van der Merwe
+# Copyright (C) 2008 Hendrik van Antwerpen
+# Copyright (C) 2008-2011, 2014-2015, 2018-2020 Philipp Wolfer
+# Copyright (C) 2009 Carlin Mangar
+# Copyright (C) 2009 Nikolai Prokoschenko
+# Copyright (C) 2011 Tim Blechmann
+# Copyright (C) 2011-2012 Chad Wilson
+# Copyright (C) 2011-2013 Michael Wiencek
+# Copyright (C) 2012 Your Name
+# Copyright (C) 2012-2013 Wieland Hoffmann
+# Copyright (C) 2013-2014, 2016, 2018-2019 Laurent Monin
+# Copyright (C) 2013-2014, 2017 Sophist-UK
+# Copyright (C) 2016 Rahul Raturi
+# Copyright (C) 2016 Simon Legner
+# Copyright (C) 2016 Suhas
+# Copyright (C) 2016-2017 Sambhav Kothari
+# Copyright (C) 2018 Vishal Choudhary
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -17,7 +36,13 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
+
+from collections import defaultdict
 from functools import partial
+from heapq import (
+    heappop,
+    heappush,
+)
 import os
 import re
 
@@ -49,12 +74,18 @@ from picard.track import (
 from picard.util import (
     encode_filename,
     icontheme,
+    natsort,
     restore_method,
 )
 
 from picard.ui.collectionmenu import CollectionMenu
+from picard.ui.colors import interface_colors
 from picard.ui.ratingwidget import RatingWidget
 from picard.ui.scriptsmenu import ScriptsMenu
+from picard.ui.widgets.tristatesortheaderview import TristateSortHeaderView
+
+
+COLUMN_ICON_SIZE = 16
 
 
 class BaseAction(QtWidgets.QAction):
@@ -73,11 +104,11 @@ class BaseAction(QtWidgets.QAction):
         raise NotImplementedError
 
 
-_album_actions = ExtensionPoint()
-_cluster_actions = ExtensionPoint()
-_clusterlist_actions = ExtensionPoint()
-_track_actions = ExtensionPoint()
-_file_actions = ExtensionPoint()
+_album_actions = ExtensionPoint(label='album_actions')
+_cluster_actions = ExtensionPoint(label='cluster_actions')
+_clusterlist_actions = ExtensionPoint(label='clusterlist_actions')
+_track_actions = ExtensionPoint(label='track_actions')
+_file_actions = ExtensionPoint(label='file_actions')
 
 
 def register_album_action(action):
@@ -119,8 +150,22 @@ class MainPanel(QtWidgets.QSplitter):
         (N_('Title'), 'title'),
         (N_('Length'), '~length'),
         (N_('Artist'), 'artist'),
+        (N_('Album Artist'), 'albumartist'),
+        (N_('Composer'), 'composer'),
+        (N_('Album'), 'album'),
+        (N_('Disc Subtitle'), 'discsubtitle'),
+        (N_('Track No.'), 'tracknumber'),
+        (N_('Disc No.'), 'discnumber'),
+        (N_('Catalog No.'), 'catalognumber'),
+        (N_('Barcode'), 'barcode'),
+        (N_('Media'), 'media'),
+        (N_('Genre'), 'genre'),
+        (N_('Fingerprint status'), '~fingerprint'),
     ]
 
+    TITLE_COLUMN = 0
+    FINGERPRINT_COLUMN = 13
+
     def __init__(self, window, parent=None):
         super().__init__(parent)
         self.window = window
@@ -136,18 +181,18 @@ class MainPanel(QtWidgets.QSplitter):
         TreeItem.text_color = self.palette().text().color()
         TreeItem.text_color_secondary = self.palette() \
             .brush(QtGui.QPalette.Disabled, QtGui.QPalette.Text).color()
-        TrackItem.track_colors = {
-            File.NORMAL: config.setting["color_saved"],
+        TrackItem.track_colors = defaultdict(lambda: TreeItem.text_color, {
+            File.NORMAL: interface_colors.get_qcolor('entity_saved'),
             File.CHANGED: TreeItem.text_color,
-            File.PENDING: config.setting["color_pending"],
-            File.ERROR: config.setting["color_error"],
-        }
-        FileItem.file_colors = {
+            File.PENDING: interface_colors.get_qcolor('entity_pending'),
+            File.ERROR: interface_colors.get_qcolor('entity_error'),
+        })
+        FileItem.file_colors = defaultdict(lambda: TreeItem.text_color, {
             File.NORMAL: TreeItem.text_color,
             File.CHANGED: TreeItem.text_color,
-            File.PENDING: config.setting["color_pending"],
-            File.ERROR: config.setting["color_error"],
-        }
+            File.PENDING: interface_colors.get_qcolor('entity_pending'),
+            File.ERROR: interface_colors.get_qcolor('entity_error'),
+        })
 
     def save_state(self):
         config.persist["splitter_state"] = self.saveState()
@@ -177,6 +222,8 @@ class MainPanel(QtWidgets.QSplitter):
         FileItem.icon_file_pending = QtGui.QIcon(":/images/file-pending.png")
         FileItem.icon_error = icontheme.lookup('dialog-error', icontheme.ICON_SIZE_MENU)
         FileItem.icon_saved = QtGui.QIcon(":/images/track-saved.png")
+        FileItem.icon_fingerprint = icontheme.lookup('fingerprint', icontheme.ICON_SIZE_MENU)
+        FileItem.icon_fingerprint_gray = icontheme.lookup('fingerprint-gray', icontheme.ICON_SIZE_MENU)
         FileItem.match_icons = [
             QtGui.QIcon(":/images/match-50.png"),
             QtGui.QIcon(":/images/match-60.png"),
@@ -238,21 +285,126 @@ class MainPanel(QtWidgets.QSplitter):
             self.update_current_view()
 
 
-class BaseTreeView(QtWidgets.QTreeWidget):
+def paint_fingerprint_icon(painter, rect, icon):
+    if not icon:
+        return
+    size = COLUMN_ICON_SIZE
+    padding_h = (rect.width() - size) / 2
+    padding_v = (rect.height() - size) / 2
+    target_rect = QtCore.QRect(rect.x() + padding_h, rect.y() + padding_v, size, size)
+    painter.drawPixmap(target_rect, icon.pixmap(size, size))
 
-    options = [
-        config.Option("setting", "color_saved", QtGui.QColor(0, 128, 0)),
-        config.Option("setting", "color_error", QtGui.QColor(200, 0, 0)),
-        config.Option("setting", "color_pending", QtGui.QColor(128, 128, 128)),
-    ]
+
+class FingerprintColumnWidget(QtWidgets.QWidget):
+    def __init__(self, file, parent=None):
+        super().__init__(parent=parent)
+        self._file = file
+
+    def paintEvent(self, event=None):
+        painter = QtGui.QPainter(self)
+        paint_fingerprint_icon(painter, self.rect(), self.decide_icon())
+
+    def decide_icon(self):
+        if getattr(self._file, 'acoustid_fingerprint', None):
+            if self.tagger.acoustidmanager.is_submitted(self._file):
+                return FileItem.icon_fingerprint_gray
+            else:
+                return FileItem.icon_fingerprint
+        else:
+            return None
+
+
+class ConfigurableColumnsHeader(TristateSortHeaderView):
+
+    def __init__(self, parent=None):
+        super().__init__(QtCore.Qt.Horizontal, parent)
+        self._visible_columns = set([0])
+
+        # The following are settings applied to default headers
+        # of QTreeView and QTreeWidget.
+        self.setSectionsMovable(True)
+        self.setStretchLastSection(True)
+        self.setDefaultAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignVCenter)
+        self.setSectionsClickable(False)
+        self.sortIndicatorChanged.connect(self.on_sort_indicator_changed)
+
+        # enable sorting, but don't actually use it by default
+        # XXX it would be nice to be able to go to the 'no sort' mode, but the
+        #     internal model that QTreeWidget uses doesn't support it
+        self.setSortIndicator(-1, QtCore.Qt.AscendingOrder)
+        self.setDefaultSectionSize(100)
+
+    def show_column(self, column, show):
+        if column == 0:  # The first column is fixed
+            return
+        self.parent().setColumnHidden(column, not show)
+        if show:
+            if self.sectionSize(column) == 0:
+                self.resizeSection(column, self.defaultSectionSize())
+            self._visible_columns.add(column)
+            if column == MainPanel.FINGERPRINT_COLUMN:
+                self.setSectionResizeMode(column, QtWidgets.QHeaderView.Fixed)
+                self.parent().resizeColumnToContents(column)
+            else:
+                self.setSectionResizeMode(column, QtWidgets.QHeaderView.Interactive)
+        elif column in self._visible_columns:
+            self._visible_columns.remove(column)
+
+    def update_visible_columns(self, columns):
+        for i, column in enumerate(MainPanel.columns):
+            self.show_column(i, i in columns)
+
+    @property
+    def visible_columns(self):
+        return self._visible_columns
+
+    def contextMenuEvent(self, event):
+        menu = QtWidgets.QMenu(self)
+
+        for i, column in enumerate(MainPanel.columns):
+            if i == 0:
+                continue
+            action = QtWidgets.QAction(_(column[0]), self.parent())
+            action.setCheckable(True)
+            action.setChecked(i in self._visible_columns)
+            action.triggered.connect(partial(self.show_column, i))
+            menu.addAction(action)
+
+        menu.addSeparator()
+        restore_action = QtWidgets.QAction(_('Restore default columns'), self.parent())
+        restore_action.triggered.connect(self.restore_defaults)
+        menu.addAction(restore_action)
+
+        menu.exec_(event.globalPos())
+        event.accept()
+
+    def restore_defaults(self):
+        self.parent().restore_default_columns()
+
+    def paintSection(self, painter, rect, index):
+        if index == MainPanel.FINGERPRINT_COLUMN:
+            painter.save()
+            super().paintSection(painter, rect, index)
+            painter.restore()
+            paint_fingerprint_icon(painter, rect, FileItem.icon_fingerprint_gray)
+        else:
+            super().paintSection(painter, rect, index)
+
+    def on_sort_indicator_changed(self, index, order):
+        if index == MainPanel.FINGERPRINT_COLUMN:
+            self.setSortIndicator(-1, QtCore.Qt.AscendingOrder)
+
+
+class BaseTreeView(QtWidgets.QTreeWidget):
 
     def __init__(self, window, parent=None):
         super().__init__(parent)
+        self.setHeader(ConfigurableColumnsHeader(self))
         self.window = window
         self.panel = parent
 
-        self.numHeaderSections = len(MainPanel.columns)
-        self.setHeaderLabels([_(h) for h, n in MainPanel.columns])
+        self.setHeaderLabels([_(h) if n != '~fingerprint' else ''
+                              for h, n in MainPanel.columns])
         self.restore_state()
 
         self.setAcceptDrops(True)
@@ -260,10 +412,6 @@ class BaseTreeView(QtWidgets.QTreeWidget):
         self.setDropIndicatorShown(True)
         self.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
 
-        # enable sorting, but don't actually use it by default
-        # XXX it would be nice to be able to go to the 'no sort' mode, but the
-        #     internal model that QTreeWidget uses doesn't support it
-        self.header().setSortIndicator(-1, QtCore.Qt.AscendingOrder)
         self.setSortingEnabled(True)
 
         self.expand_all_action = QtWidgets.QAction(_("&Expand all"), self)
@@ -294,6 +442,8 @@ class BaseTreeView(QtWidgets.QTreeWidget):
                 menu.addAction(self.window.track_search_action)
                 plugin_actions.extend(_file_actions)
             menu.addAction(self.window.browser_lookup_action)
+            if obj.num_linked_files > 0:
+                menu.addAction(self.window.generate_fingerprints_action)
             menu.addSeparator()
             if isinstance(obj, NonAlbumTrack):
                 menu.addAction(self.window.refresh_action)
@@ -308,10 +458,12 @@ class BaseTreeView(QtWidgets.QTreeWidget):
                 menu.addAction(self.window.cluster_action)
             else:
                 menu.addAction(self.window.album_search_action)
+            menu.addAction(self.window.generate_fingerprints_action)
             plugin_actions = list(_cluster_actions)
         elif isinstance(obj, ClusterList):
             menu.addAction(self.window.autotag_action)
             menu.addAction(self.window.analyze_action)
+            menu.addAction(self.window.generate_fingerprints_action)
             plugin_actions = list(_clusterlist_actions)
         elif isinstance(obj, File):
             if can_view_info:
@@ -323,11 +475,14 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             menu.addAction(self.window.autotag_action)
             menu.addAction(self.window.analyze_action)
             menu.addAction(self.window.track_search_action)
+            menu.addAction(self.window.generate_fingerprints_action)
             plugin_actions = list(_file_actions)
         elif isinstance(obj, Album):
             if can_view_info:
                 menu.addAction(self.window.view_info_action)
             menu.addAction(self.window.browser_lookup_action)
+            if obj.get_num_total_files() > 0:
+                menu.addAction(self.window.generate_fingerprints_action)
             menu.addSeparator()
             menu.addAction(self.window.refresh_action)
             plugin_actions = list(_album_actions)
@@ -345,7 +500,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             loading.setDisabled(True)
             bottom_separator = True
 
-            if len(self.selectedIndexes()) == len(MainPanel.columns):
+            if len(self.selectedItems()) == 1:
                 def _add_other_versions():
                     releases_menu.removeAction(loading)
                     heading = releases_menu.addAction(obj.release_group.version_headings)
@@ -356,34 +511,44 @@ class BaseTreeView(QtWidgets.QTreeWidget):
 
                     versions = obj.release_group.versions
 
-                    albumtracks = obj.get_num_total_files() if obj.get_num_total_files() else len(obj.tracks)
+                    album_tracks_count = obj.get_num_total_files() or len(obj.tracks)
                     preferred_countries = set(config.setting["preferred_release_countries"])
                     preferred_formats = set(config.setting["preferred_release_formats"])
-                    matches = ("trackmatch", "countrymatch", "formatmatch")
-                    priorities = {}
-                    for version in versions:
-                        priority = {
-                            "trackmatch": "0" if version['totaltracks'] == albumtracks else "?",
-                            "countrymatch": "0" if len(preferred_countries) == 0 or preferred_countries & set(version['countries'] or '') else "?",
-                            "formatmatch": "0" if len(preferred_formats) == 0 or preferred_formats & set(version['formats'] or '') else "?",
-                        }
-                        priorities[version['id']] = "".join(priority[k] for k in matches)
-                    versions.sort(key=lambda version: priorities[version['id']] + version['name'])
+                    ORDER_BEFORE, ORDER_AFTER = 0, 1
 
-                    priority = normal = False
+                    alternatives = []
                     for version in versions:
-                        if not normal and "?" in priorities[version['id']]:
-                            if priority:
+                        trackmatch = countrymatch = formatmatch = ORDER_BEFORE
+                        if version['totaltracks'] != album_tracks_count:
+                            trackmatch = ORDER_AFTER
+                        if preferred_countries:
+                            countries = set(version['countries'])
+                            if not countries or not countries.intersection(preferred_countries):
+                                countrymatch = ORDER_AFTER
+                        if preferred_formats:
+                            formats = set(version['formats'])
+                            if not formats or not formats.intersection(preferred_formats):
+                                formatmatch = ORDER_AFTER
+                        group = (trackmatch, countrymatch, formatmatch)
+                        # order by group, name, and id on push
+                        heappush(alternatives, (group, version['name'], version['id']))
+
+                    prev_group = None
+                    while alternatives:
+                        group, action_text, release_id = heappop(alternatives)
+                        if group != prev_group:
+                            if prev_group is not None:
                                 releases_menu.addSeparator()
-                            normal = True
-                        else:
-                            priority = True
-                        action = releases_menu.addAction(version["name"])
+                            prev_group = group
+                        action = releases_menu.addAction(action_text)
                         action.setCheckable(True)
-                        if obj.id == version["id"]:
+                        if obj.id == release_id:
                             action.setChecked(True)
-                        action.triggered.connect(partial(obj.switch_release_version, version["id"]))
+                        action.triggered.connect(partial(obj.switch_release_version, release_id))
 
+                    versions_count = len(versions)
+                    if versions_count > 1:
+                        releases_menu.setTitle(_("&Other versions (%d)") % versions_count)
                 if obj.release_group.loaded:
                     _add_other_versions()
                 else:
@@ -402,6 +567,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
 
         # Using type here is intentional. isinstance will return true for the
         # NatAlbum instance, which can't be part of a collection.
+        # pylint: disable=C0123
         selected_albums = [a for a in self.window.selected_objects if type(a) == Album]
         if selected_albums:
             if not bottom_separator:
@@ -445,19 +611,25 @@ class BaseTreeView(QtWidgets.QTreeWidget):
 
     @restore_method
     def restore_state(self):
-        sizes = config.persist[self.view_sizes.name]
-        header = self.header()
-        sizes = sizes.split(" ")
-        try:
-            for i in range(self.numHeaderSections - 1):
-                header.resizeSection(i, int(sizes[i]))
-        except IndexError:
-            pass
+        self._restore_state(config.persist[self.header_state.name])
 
     def save_state(self):
-        cols = range(self.numHeaderSections - 1)
-        sizes = " ".join(str(self.header().sectionSize(i)) for i in cols)
-        config.persist[self.view_sizes.name] = sizes
+        config.persist[self.header_state.name] = self.header().saveState()
+
+    def restore_default_columns(self):
+        self._restore_state(None)
+
+    def _restore_state(self, header_state):
+        header = self.header()
+        if header_state:
+            header.restoreState(header_state)
+            for i in range(0, self.columnCount()):
+                header.show_column(i, not self.isColumnHidden(i))
+        else:
+            header.update_visible_columns([0, 1, 2])
+            for i, size in enumerate([250, 50, 100]):
+                header.resizeSection(i, size)
+            self.sortByColumn(-1, QtCore.Qt.AscendingOrder)
 
     def supportedDropActions(self):
         return QtCore.Qt.CopyAction | QtCore.Qt.MoveAction
@@ -503,6 +675,14 @@ class BaseTreeView(QtWidgets.QTreeWidget):
             mimeData.setUrls(files)
         return mimeData
 
+    def scrollTo(self, index, scrolltype=QtWidgets.QAbstractItemView.EnsureVisible):
+        # QTreeView.scrollTo resets the horizontal scroll position to 0.
+        # Reimplemented to maintain current horizontal scroll position.
+        hscrollbar = self.horizontalScrollBar()
+        xpos = hscrollbar.value()
+        super().scrollTo(index, scrolltype)
+        hscrollbar.setValue(xpos)
+
     @staticmethod
     def drop_urls(urls, target):
         files = []
@@ -545,20 +725,23 @@ class BaseTreeView(QtWidgets.QTreeWidget):
                 item = parent.child(index)
             if item is not None:
                 target = item.obj
+        if isinstance(self, FileTreeView) and target is None:
+            target = self.tagger.unclustered_files
         log.debug("Drop target = %r", target)
         handled = False
         # text/uri-list
         urls = data.urls()
         if urls:
-            self.drop_urls(urls, target)
+            # Use QTimer.singleShot to run expensive processing outside of the drop handler.
+            QtCore.QTimer.singleShot(0, partial(self.drop_urls, urls, target))
             handled = True
         # application/picard.album-list
         albums = data.data("application/picard.album-list")
         if albums:
-            if isinstance(self, FileTreeView) and target is None:
-                target = self.tagger.unclustered_files
             albums = [self.tagger.load_album(id) for id in bytes(albums).decode().split("\n")]
-            self.tagger.move_files(self.tagger.get_files_from_objects(albums), target)
+            # Use QTimer.singleShot to run expensive processing outside of the drop handler.
+            QtCore.QTimer.singleShot(0, partial(self.tagger.move_files,
+                self.tagger.get_files_from_objects(albums), target))
             handled = True
         return handled
 
@@ -589,7 +772,7 @@ class BaseTreeView(QtWidgets.QTreeWidget):
 
 class FileTreeView(BaseTreeView):
 
-    view_sizes = config.TextOption("persist", "file_view_sizes", "250 40 100")
+    header_state = config.Option("persist", "file_view_header_state", QtCore.QByteArray())
 
     def __init__(self, window, parent=None):
         super().__init__(window, parent)
@@ -614,12 +797,12 @@ class FileTreeView(BaseTreeView):
         self.set_clusters_text()
 
     def set_clusters_text(self):
-        self.clusters.setText(0, '%s (%d)' % (_("Clusters"), len(self.tagger.clusters)))
+        self.clusters.setText(MainPanel.TITLE_COLUMN, '%s (%d)' % (_("Clusters"), len(self.tagger.clusters)))
 
 
 class AlbumTreeView(BaseTreeView):
 
-    view_sizes = config.TextOption("persist", "album_view_sizes", "250 40 100")
+    header_state = config.Option("persist", "album_view_header_state", QtCore.QByteArray())
 
     def __init__(self, window, parent=None):
         super().__init__(window, parent)
@@ -629,8 +812,12 @@ class AlbumTreeView(BaseTreeView):
         self.tagger.album_removed.connect(self.remove_album)
 
     def add_album(self, album):
-        item = AlbumItem(album, True, self)
-        item.setIcon(0, AlbumItem.icon_cd)
+        if isinstance(album, NatAlbum):
+            item = NatAlbumItem(album, True)
+            self.insertTopLevelItem(0, item)
+        else:
+            item = AlbumItem(album, True, self)
+        item.setIcon(MainPanel.TITLE_COLUMN, AlbumItem.icon_cd)
         for i, column in enumerate(MainPanel.columns):
             font = item.font(i)
             font.setBold(True)
@@ -652,21 +839,26 @@ class TreeItem(QtWidgets.QTreeWidgetItem):
             obj.item = self
         self.sortable = sortable
         self.setTextAlignment(1, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
+        self.setTextAlignment(7, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
+        self.setTextAlignment(8, QtCore.Qt.AlignRight | QtCore.Qt.AlignVCenter)
 
     def __lt__(self, other):
         if not self.sortable:
             return False
         column = self.treeWidget().sortColumn()
+        return self.sortkey(column) < other.sortkey(column)
+
+    def sortkey(self, column):
         if column == 1:
-            return (self.obj.metadata.length or 0) < (other.obj.metadata.length or 0)
-        return self.text(column).lower() < other.text(column).lower()
+            return self.obj.metadata.length or 0
+        return natsort.natkey(self.text(column).lower())
 
 
 class ClusterItem(TreeItem):
 
     def __init__(self, *args):
         super().__init__(*args)
-        self.setIcon(0, ClusterItem.icon_dir)
+        self.setIcon(MainPanel.TITLE_COLUMN, ClusterItem.icon_dir)
 
     def update(self):
         for i, column in enumerate(MainPanel.columns):
@@ -704,18 +896,21 @@ class AlbumItem(TreeItem):
 
     def update(self, update_tracks=True):
         album = self.obj
+        selection_changed = self.isSelected()
         if update_tracks:
             oldnum = self.childCount() - 1
             newnum = len(album.tracks)
             if oldnum > newnum:  # remove old items
                 for i in range(oldnum - newnum):
+                    item = self.child(newnum)
+                    selection_changed |= item.isSelected()
                     self.takeChild(newnum)
                 oldnum = newnum
             # update existing items
             for i in range(oldnum):
                 item = self.child(i)
-                item.setSelected(False)
                 track = album.tracks[i]
+                selection_changed |= item.isSelected() and item.obj != track
                 item.obj = track
                 track.item = item
                 item.update(update_album=False)
@@ -729,43 +924,68 @@ class AlbumItem(TreeItem):
                 for item in items:  # Update after insertChildren so that setExpanded works
                     item.update(update_album=False)
         if album.errors:
-            self.setIcon(0, AlbumItem.icon_error)
-            self.setToolTip(0, _("Error"))
+            self.setIcon(MainPanel.TITLE_COLUMN, AlbumItem.icon_error)
+            self.setToolTip(MainPanel.TITLE_COLUMN, _("Error"))
         elif album.is_complete():
             if album.is_modified():
-                self.setIcon(0, AlbumItem.icon_cd_saved_modified)
-                self.setToolTip(0, _("Album modified and complete"))
+                self.setIcon(MainPanel.TITLE_COLUMN, AlbumItem.icon_cd_saved_modified)
+                self.setToolTip(MainPanel.TITLE_COLUMN, _("Album modified and complete"))
             else:
-                self.setIcon(0, AlbumItem.icon_cd_saved)
-                self.setToolTip(0, _("Album unchanged and complete"))
+                self.setIcon(MainPanel.TITLE_COLUMN, AlbumItem.icon_cd_saved)
+                self.setToolTip(MainPanel.TITLE_COLUMN, _("Album unchanged and complete"))
         else:
             if album.is_modified():
-                self.setIcon(0, AlbumItem.icon_cd_modified)
-                self.setToolTip(0, _("Album modified"))
+                self.setIcon(MainPanel.TITLE_COLUMN, AlbumItem.icon_cd_modified)
+                self.setToolTip(MainPanel.TITLE_COLUMN, _("Album modified"))
             else:
-                self.setIcon(0, AlbumItem.icon_cd)
-                self.setToolTip(0, _("Album unchanged"))
+                self.setIcon(MainPanel.TITLE_COLUMN, AlbumItem.icon_cd)
+                self.setToolTip(MainPanel.TITLE_COLUMN, _("Album unchanged"))
         for i, column in enumerate(MainPanel.columns):
             self.setText(i, album.column(column[1]))
-        if self.isSelected():
-            TreeItem.window.update_selection()
+        if selection_changed:
+            TreeItem.window.panel.update_current_view()
+        # Workaround for PICARD-1446: Expand/collapse indicator for the release
+        # is briefly missing on Windows
+        self.emitDataChanged()
+
+    def __lt__(self, other):
+        # Always show NAT entry on top, see also NatAlbumItem.__lt__
+        if isinstance(other, NatAlbumItem):
+            return not other.__lt__(self)
+        return super().__lt__(other)
+
+
+class NatAlbumItem(AlbumItem):
+    def __lt__(self, other):
+        # Always show NAT entry on top
+        order = self.treeWidget().header().sortIndicatorOrder()
+        return order == QtCore.Qt.AscendingOrder
 
 
 class TrackItem(TreeItem):
 
     def update(self, update_album=True, update_files=True):
         track = self.obj
+        tree_widget = self.treeWidget()
         if track.num_linked_files == 1:
             file = track.linked_files[0]
             file.item = self
             color = TrackItem.track_colors[file.state]
             bgcolor = get_match_color(file.similarity, TreeItem.base_color)
             icon = FileItem.decide_file_icon(file)
-            self.setToolTip(0, _(FileItem.decide_file_icon_info(file)))
+            self.setToolTip(MainPanel.TITLE_COLUMN, _(FileItem.decide_file_icon_info(file)))
             self.takeChildren()
             self.setExpanded(False)
+            self.setToolTip(MainPanel.FINGERPRINT_COLUMN, FileItem.decide_fingerprint_icon_info(file))
+            if tree_widget:
+                if not tree_widget.itemWidget(self, MainPanel.FINGERPRINT_COLUMN):
+                    tree_widget.setItemWidget(self, MainPanel.FINGERPRINT_COLUMN,
+                        FingerprintColumnWidget(file=file))
         else:
-            self.setToolTip(0, "")
+            self.setToolTip(MainPanel.TITLE_COLUMN, "")
+            self.setToolTip(MainPanel.FINGERPRINT_COLUMN, "")
+            if tree_widget:
+                tree_widget.setItemWidget(self, MainPanel.FINGERPRINT_COLUMN, None)
             if track.ignored_for_completeness():
                 color = TreeItem.text_color_secondary
             else:
@@ -799,10 +1019,10 @@ class TrackItem(TreeItem):
                     self.addChildren(items)
             self.setExpanded(True)
         if track.error:
-            self.setIcon(0, TrackItem.icon_error)
-            self.setToolTip(0, track.error)
+            self.setIcon(MainPanel.TITLE_COLUMN, TrackItem.icon_error)
+            self.setToolTip(MainPanel.TITLE_COLUMN, track.error)
         else:
-            self.setIcon(0, icon)
+            self.setIcon(MainPanel.TITLE_COLUMN, icon)
         for i, column in enumerate(MainPanel.columns):
             self.setText(i, track.column(column[1]))
             self.setForeground(i, color)
@@ -817,13 +1037,19 @@ class FileItem(TreeItem):
 
     def update(self, update_track=True):
         file = self.obj
-        self.setIcon(0, FileItem.decide_file_icon(file))
+        self.setIcon(MainPanel.TITLE_COLUMN, FileItem.decide_file_icon(file))
+        self.setToolTip(MainPanel.FINGERPRINT_COLUMN, self.decide_fingerprint_icon_info(file))
         color = FileItem.file_colors[file.state]
         bgcolor = get_match_color(file.similarity, TreeItem.base_color)
         for i, column in enumerate(MainPanel.columns):
             self.setText(i, file.column(column[1]))
             self.setForeground(i, color)
             self.setBackground(i, bgcolor)
+        tree_widget = self.treeWidget()
+        if tree_widget:
+            if not tree_widget.itemWidget(self, MainPanel.FINGERPRINT_COLUMN):
+                tree_widget.setItemWidget(self, MainPanel.FINGERPRINT_COLUMN,
+                    FingerprintColumnWidget(file=file))
         if self.isSelected():
             TreeItem.window.update_selection()
 
@@ -859,3 +1085,13 @@ class FileItem(TreeItem):
                 return FileItem.match_icons_info[int(file.similarity * 5 + 0.5)]
         elif file.state == File.PENDING:
             return N_("Pending")
+
+    @staticmethod
+    def decide_fingerprint_icon_info(file):
+        if file.acoustid_fingerprint:
+            if QtCore.QObject.tagger.acoustidmanager.is_submitted(file):
+                return _('Fingerprint has already been submitted')
+            else:
+                return _('Unsubmitted fingerprint')
+        else:
+            return _('No fingerprint was calculated for this file, use "Scan" or "Generate AcoustID Fingerprints" to calculate the fingerprint.')
